/*!
 * RedScript compiler
 * Copyright(c) 2013 Adam Brodzinski <adambrodzinski@gmail.com>
 * MIT Licensed
 */

String.prototype.lastChar = function(){
 return this[this.length - 1];
};

module.exports = function(file) {
    var lines = file.split('\n')
      , debug = false;


    /**
     * Checks to see if keyword is used inside a string
     * @param  {string} match keyword to match
     * @param  {string} ln    line to match against
     * @return {bool}         true if match is found
     */
    function insideString(match, ln) {
        var regex = "['\"][\\w\\s]*[^\\w]" + match + "[^\\w][\\w\\s]*['\"]";
        var insideStr = new RegExp(regex, "g");
        return insideStr.test(ln);
    }

    // Iterate through each line and transform text
    lines = lines.map(function(line, index) {
        if (debug) console.log(index + "  " + line);

        // Matches an @foo or @. or @
        var atSymbol = /@(?=[\w$])|@\.|@/g
        /**
         * Matches `func` keyword with and without parens
         * @type {RegExp}
         * given `func foo(p)` & `func bar`
         * $1 captures foo
         * $2 captures (p)
         * $3 captures bar     (matches only no parens)
         */
          , funcRgx = /func\s([\w$]+)(\((?:[\w$, ]+)?\))|func\s([\w$]+)/g

        /**
         * Matches `when` OR `when` and `then`.
         * @type {RegExp} See function for more info
         */
          , switchWhenThen = /(when)\s(.+)(then)(.+)|(when)\s(.+)/


        /**
         * Matches string interpolation inside double quotes
         * $1 captures left of `#{`, inc double quote
         * $3 captures interpolation inside of `#{  }`
         * $5 captures right of `}`, inc double quote
         * @type {RegExp}
         */
          , strInt = /(".*)(#\{)([^\{\}]+)(\})(.*")/

        /**
         * Matches an anonymous function 'block'
         * $0 examp: `on(foo, do |x,y|`
         * $1 `on(foo, `
         * $2 `x,y`
         * @type {RegExp}
         */
          , anonBlock = /(\(.*)\bdo\b(?:\s\|([\w_$$,\s]*)\|)?/

        /**
         * Matches a RedScript object literal declaration
         * $0 `object foo inherits bar`
         * $1 `foo`
         * $2 `bar`
         */
          , objLit = /object ([\$\w]+)(?: inherits ([\$\w]+))?/

        /**
         * Matches a `def` inside of an object literal
         * $0 `def foo(p1)
         * $1 `foo`
         * $2 `(p1)`
         */
          , objLitDef = /\bdef\b ([\$\w]+)(\([\$\w, ]*\))?/

        /**
         * Matches `def foo.bar` methods
         * $0 matches `def foo.bar(p1)`
         * $1 matches `foo.bar`
         * $2 matches `(p1)`
         */
          , dotDef = /\bdef\b ([\$\w]+\.[\$\w]+)(\([\$\w, ]*\))?/

         /**
          * Matches `def` methods attached to prototype
          * $0 `def foo >>> bar(p1)
          * $1 `foo`
          * $2 `bar`
          * $3 `(p1)`
          */
          , protoDef = /\bdef\b ([\$\w]+) >>> ([\$\w]+)(\([\$\w, ]*\))?/;

        // normalize comments before processing // ******* buggy needs fixed *******///
        line = line.replace(/#(?![\{\w])/g, "//");

        // if line begins with a comment, return
        if (/^\/\//.test(line.trimLeft())) return line;

        /**
         * Alias `do` with `{`
         * fakes negative lookbehind for { or word char, this
         * prevents aliasing a properly formated do loop
         * @return {string} returns `}` if lookbehind is false
         */
        /*
         *line = line.replace(/(\w)?do(?! \{|\{|\w)/g, function($0, $1) {
         *    return $1 ? $0 : '{';
         *});
         */


        /**
         * Alias `end-` with `end);` or `});`
         */
        line = line.replace(/\bend\b-/g, '});');

        /**
         * Alias `end` with `}`
         * fakes a negative lookbehind for a word char
         * @return {string} returns `}` if lookbehind is false
         */
        line = line.replace(/(\w)?end(?!\w)/g, function($0, $1) {
            return $1 ? $0 : '}';
        });


        /**
         * Find any instance of @ and replace it with the
         * appropriate this.
         */
        line = line.replace(atSymbol, function(match) {
            // if line contains @foo, @_foo or @$foo
            if (/@(?=[\w$])/.test(line)) return 'this.';
            // else if matches @.foo, @._foo or @.$foo
            else if (match === '@.') return "this.";
            else return "this";
        });


        /**
         * Alias puts with console.log
         * if it's bracketless, insert bracket and close at EOL
         */
        line = line.replace(/puts\(/, "console.log(");
         // if bracketless puts, replace puts & append );
        if (/puts /.test(line)) {
            line = line.replace(/puts /, "console.log(") + ");";
        }


        /**
         * Alias printf with process.stdout.write
         * if it's bracketless, insert bracket and close EOL
         */
        line = line.replace(/printf\(/, "process.stdout.write(");
         // if bracketless puts, replace puts & append );
        if (/printf /.test(line)) {
            line = line.replace(/printf /, "process.stdout.write(") + ");";
        }


        /**
         * Replace an arrow function to regular function. Function can either
         * be a single arrow `->`, with empty parens `() ->` and an arrow with
         * params, `(foo, bar) ->. One space between arrow is optional.
         *
         * @param  {string} $0  Entire match
         * @param  {string} $1  matches outside of parens
         * @param  {string} $2  matches inside of parens
         * @return {string}     a normal function `function () {`
         */
        line = line.replace(/(\(([\w$, ]*?)\))?\s{0,1}->/, function($0, $1, $2) {
            // if there are no parens in expression
            if (!$1) {
                return "function () {";
            // else if it has parens and params
            } else if ($1 && $2) {
                return "function (" + $2 + ") {" ;
            // else if it parens with no params
            } else if ($1) {
                return "function () {";
            } else {
                throw console.log("Unexpected arrow formatting. Saw: ", $0);
            }
        });


        /**
         * Replaces `func foo` declaration with `var foo = function() {`
         * Parens are optional with no params are use.
         *
         * @param  {regex}  funcRgx - Regex declared top of scope
         * @param  {string} $0     Entire match
         * @param  {string} $1     Function name
         * @param  {string} $2     Function name if no parens are found
         * @param  {string} $3)    Parens including any params
         * @return {string}        Processed function string
         */
        line = line.replace(funcRgx, function($0, $1, $2, $3) {
            var nameP = $1
              , nameNP = $3
              , parens = $2;

            if (nameP && parens) {
                return "var " + nameP +" = function" + parens + " {";
            } else if (nameNP) {
                return "var " + nameNP +" = function" + "() {";
            }
        });


        /**
         * Inserts brackets into switch statement
         * @param  {string} $0  Entire match
         * @param  {string} $1  Matches switch
         * @param  {string} $2  Matches expression
         * @return {string}     Processed switch line
         */
        line = line.replace(/(switch) (.+)/, function($0, $1, $2) {
            return $1 + " (" + $2 + ") {";
        });


        /**
         * Replaces when with case. If then is found a one line
         * statement is assumed. This currently inserts a semicolon to prevent
         * errors with appended break.
         * @param  {string} $0  Entire match
         * @param  {string} $1  Matches `when`
         * @param  {string} $2  Label
         * @param  {string} $3  Matches `then`
         * @param  {string} $4  clause (after then)
         *                  OR
         * @param  {string} $5  Matches `when`
         * @param  {string} $6  Label
         * @return {string}     Processed when line
         */
        line = line.replace(switchWhenThen, function($0, $1 ,$2 ,$3 ,$4 ,$5 ,$6) {
            // has `when` and `then` one liner
            if ($1 && $3) {
                return "case " + $2 + ":" + $4 + ' ; break;';
            // line just has `when`
            } else if ($5) {
                return "case " + $6 + ":";
            }
        });


        /**
         * Replaces default with default: in a switch statement
         */
        line = line.replace(/\bdefault\b(?!\:)/, 'default:');


        /**
         * Replaces an `else if` with `} else if () {`
         * @param  {string} $0  Entire match
         * @param  {string} $1  Matches `condition`
         */
        line = line.replace(/else if (?!\s*\()(.+)/, function($0, $1) {
            // if `else if` is inside a string, bail out
            if (insideString("else if", line)) return $0;
            return '} else if (' + $1 + ') {';
        });


        /**
         * Replaces an `else` with `} else {`
         * @param  {string} $0  Entire match
         */
        line = line.replace(/else(?!(?:\s*\{)| if)/, function($0) {
            // if `else` is inside a string, bail out
            if (insideString("else", line)) return $0;
            return '} else {';
        });


        /**
         * Replaces an `if condition` with `if (condition) {`
         * @param  {string} $0  Entire match
         * @param  {string} $1  Matches `condition`
         */
        line = line.replace(/if (?!\s*\()(.+)/, function($0, $1) {
            // if `if` is inside a string, bail out
            if (insideString("if", line)) return $0;
            return 'if (' + $1 + ') {';
        });

        /**
         * Replaces text inside string interpolation brackets `#{ }`
         * @return {string} transformed string
         */
        line = line.replace(strInt, function($0, $1 ,$2 ,$3 ,$4 ,$5) {
            var left = $1   // left of `#{`, inc double quote
              , right = $5  // right of `}`, inc double quote
              , middle = $3; // inside of `#{  }`

            // if middle contains scary chars like +*-% etc, wrap in parens
            if (/[^\w\s$$]/.test(middle)) {
                middle = '(' + middle + ')';
            }
            // if interp is to the far left/right, don't add extra quotes
            if (left === '"')  return middle + ' + "' + right;
            if (right === '"') return left + '" + ' + middle;
            
            return left + '" + ' + middle + ' + "' + right;
        });


        /**
         * Replaces an anonymous function block
         * Matches an anonymous function 'block'
         * $0 examp: `on(foo, do |x,y|`
         * $1 `on(foo, `
         * $2 `x,y`
         * @return {string} transformed funciton
         */
        line = line.replace(anonBlock, function($0, $1, $2) {
            var $1tr = $1.trimRight();
            var insertComma = /\(.*[^,]$/.test($1tr);
            // if line has params, insert them
            if ($2) {
                // if comma wasn't inserted between block and params
                if ( insertComma ) {
                    return $1 + ', ' + 'function(' + $2 + ') {';
                }
                return $1 + 'function(' + $2 + ') {';
            // else no params
            } else {
                // if comma wasn't inserted between block and params
                if ( insertComma ) {
                    return $1 + ', ' + 'function() {';
                }
                return $1 + 'function() {';
            }
        });


        /**
         * Replaces object literal syntax with vanilla
         * Optional inherits keyword connects __proto__
         * $0 matches `object foo inherits bar`
         * $1 matches `foo`
         * $2 matches `bar`
         */
        line = line.replace(objLit, function($0, $1, $2) {
            if (insideString("object", line)) return $0;
            return 'var ' + $1 + ' = {';
        });


        /**
         * Replaces `def foo.bar` methods with vanilla methods
         * $0 matches `def foo.bar(p1)`
         * $1 matches `foo.bar`
         * $2 matches `(p1)`
         */
        line = line.replace(dotDef, function($0, $1, $2) {
            if (insideString("def", line)) return $0;
            // if def declaration has parens
            if ($2)
                return $1 + ' = function' + $2 + ' {';
            else
                return $1 + ' = function() {';
        });


        /**
         * Replaces `def foo >>> bar` methods with vanilla methods
         * $0 matches `def foo >>> bar(p1)`
         * $1 matches `foo`
         * $2 matches `bar`
         * $3 matches `(p1)
         */
        line = line.replace(protoDef, function($0, $1, $2, $3) {
            if (insideString("def", line)) return $0;
            // if def declaration has parens
            if ($3)
                return $1 + '.prototype.' + $2 + ' = function' + $3 + ' {';
            else
                return $1 + '.prototype.' + $2 + ' = function() {';
        });


        /**
         * Replaces `def` methods with vanilla methods inside
         * of an object literal.
         * $0 matches `def foo(p1)`
         * $1 matches `foo`
         * $2 matches `(p1)`
         */
        line = line.replace(objLitDef, function($0, $1, $2) {
            if (insideString("def", line)) return $0;
            // if def declaration has parens
            if ($2)
              return $1 + ': ' + 'function' + $2 + ' {';
            else
              return $1 + ': ' + 'function() {';
        });


        /**
         * Conditional assignment operator
         * $0 matches `foo =|| bar`
         * $1 matches `foo`
         * $2 matches `bar`
         * @return {String} transformed line
         */
        var condAssignment = /([\w\$]+)\s*\|\|=\s*(.+)/;
        line = line.replace(condAssignment, function($0, $1, $2) {
            //if (insideString("||=", line)) return $0;
            return $1 + ' = ' + $1 + ' || ' + $2;
        });


        if (debug) console.log(index + "  " + line);
        return line;
    });
    return lines.join('\n');
};
